﻿using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using UnityEngine.EventSystems;

/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea
{
    public class AdSorbData
    {
        public DAZBone dazBome;
        public Vector3 point;
        public Vector3 normal;
    }

    public struct AdSorbMode
    {
        public const int None = 0;
        public const int LinkUI = 1;
        public const int LinkMouse = 2;
    }

    public class AdsorbModule : EasyMorphyModuleBase
    {

        private JSONStorableBool isPriorityPrimaryNode;

        private GLDraw gldraw;
        private AdSorbData adsorbData;
        private bool isAdsorbActive;
        private Camera targetCamera;
        private bool isAdsorbLocked;

        public bool isDrawMarkPoint = true;

        public int currentAdsorbMode = AdSorbMode.None;

        public bool IsAdsorbActive
        {
            get { return isAdsorbActive; }
            set { isAdsorbActive = value; }
        }

        public bool IsAdsorbLocked
        {
            get { return this.isAdsorbLocked; }
            set { isAdsorbLocked = value; }
        }

        public AdSorbData AdSorbData
        {
            get {
                if (isAdsorbActive)
                {
                    return this.adsorbData;
                }

                SuperController.LogError("can't adsorb!");
                return null;
            }
        }

        // 吸附功能连线结束点
        private Vector3 LinkEndPoint
        {
            get {
                if (currentAdsorbMode == AdSorbMode.LinkMouse)
                {
                    return Input.mousePosition;
                }
                else if (currentAdsorbMode == AdSorbMode.LinkUI || isAdsorbLocked)
                {
                    if (ower.CurrentFloatPanel != null)
                    {
                        return targetCamera.WorldToScreenPoint(ower.CurrentFloatPanel.transform.position);
                    }
                }

                return Vector3.zero;
            }
        }

        public AdsorbModule(EasyMorphy ower)
        {
            this.ower = ower;
            this.gldraw = new GLDraw();
            this.targetCamera = SuperController.singleton.MonitorCenterCamera;
        }

        public override void Init()
        {
            InitUI();
        }

        private void InitUI()
        {
            isPriorityPrimaryNode = Utils.SetupToggle(ower, "IsPriorityPrimaryNode", true, false);
        }

        public override void OnUpdate()
        {
            if (currentAdsorbMode == AdSorbMode.None || isAdsorbLocked) return;

            var bones = ower.BonesList;
            if (bones.Count > 0)
            {
                var mainClosestBone = ower.MainBonesList[0];
                float mainClosestDistance = Mathf.Infinity;
                float cloestDistance = Mathf.Infinity;
                for (int i = 0; i < ower.MainBonesList.Count; i++)
                {
                    float disMin = Vector3.Distance(targetCamera.WorldToScreenPoint(mainClosestBone.transform.position), LinkEndPoint);
                    float disCur = Vector3.Distance(targetCamera.WorldToScreenPoint(ower.MainBonesList[i].transform.position), LinkEndPoint);

                    if (disMin > disCur)
                    {
                        mainClosestDistance = disCur;
                        mainClosestBone = ower.MainBonesList[i];
                    }
                }

                var closestBone = bones[0];
                for (int i = 0; i < bones.Count; i++)
                {
                    float disMin = Vector3.Distance(targetCamera.WorldToScreenPoint(closestBone.transform.position), LinkEndPoint);
                    float disCur = Vector3.Distance(targetCamera.WorldToScreenPoint(bones[i].transform.position), LinkEndPoint);

                    if (disMin > disCur)
                    {
                        cloestDistance = disCur;
                        closestBone = bones[i];
                    }
                }

                var finalClosetBone = (!isPriorityPrimaryNode.val || cloestDistance / mainClosestDistance < 0.5f) ? closestBone : mainClosestBone;

                adsorbData = new AdSorbData() { dazBome = finalClosetBone, normal = Vector3.up, point = finalClosetBone.transform.position };
            }
        }

        public override void OnPostRender()
        {
            if (isAdsorbActive)
            {
                var screenPoint = targetCamera.WorldToScreenPoint(this.adsorbData.point);
                Vector2 from = new Vector2() { x = screenPoint.x - 10, y = screenPoint.y };
                Vector2 to = new Vector2() { x = screenPoint.x + 10, y = screenPoint.y };
                Vector2 from1 = new Vector2() { x = screenPoint.x, y = screenPoint.y - 10 };
                Vector2 to1 = new Vector2() { x = screenPoint.x, y = screenPoint.y + 10 };

                // shape x
                if (this.currentAdsorbMode == AdSorbMode.LinkMouse)
                {
                    gldraw.DrawLine2D(new List<GLLine2D> { new GLLine2D() { from = from, to = to }, new GLLine2D() { from = from1, to = to1 } }, Color.cyan);

                    // draw mark squa
                    DrawActiveBoneMarkPoint();
                }

                // line
                if (!EventSystem.current.IsPointerOverGameObject() || isAdsorbLocked)
                {
                    gldraw.DrawLine2D(new List<GLLine2D>() { new GLLine2D() { from = screenPoint, to = LinkEndPoint } }, Color.gray);
                }
            }
        }

        private void DrawActiveBoneMarkPoint()
        {
            if (isDrawMarkPoint)
            {
                List<DAZBone> activeBones = ower.MorphyActiveBoneList;
                if (activeBones.Count > 0)
                {
                    for (int i = 0; i < activeBones.Count; i++)
                    {
                        Vector3 targetObjectPosition = activeBones[i].transform.position;
                        Camera monitorCenterCamera = SuperController.singleton.MonitorCenterCamera;
                        float distanceObjectToCamera = Vector3.Distance(targetObjectPosition, monitorCenterCamera.transform.position);
                        var anchorPosition = monitorCenterCamera.WorldToScreenPoint(targetObjectPosition);
                        Rect rect = new Rect(anchorPosition.x, anchorPosition.y, 10 / distanceObjectToCamera, 10 / distanceObjectToCamera);
                        gldraw.DrawQuad2D(rect, Color.yellow);
                    }
                }
            }
        }

        public override void UnInit()
        {
            
        }
    }
}
