﻿using UnityEngine;
using System.Collections.Generic;
using UnityEngine.UI;
using UnityEngine.EventSystems;
using UnityEngine.Events;
using MeshVR;
using System;
using System.Linq;
using System.Text;
using SimpleJSON;
using MVR.FileManagementSecure;

/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea
{

    public class EasyMorphy : MVRScript
    {

        struct GameMode
        {
            public const int VR = 0;
            public const int Destop = 1;
        }

        struct OperateState
        {
            public const int None = 0;
            public const int AligningBoneNode = 1;
            public const int Linking = 2;
        }

        class BoneMophyConfig
        {
            public string boneName;
            public List<string> uids = new List<string>();
        }

        private const string presetSavePath = "Saves\\easyMorph";
        private const string presetEndfix = "morph";

        private int currentGameMode;
        private int currentOperateState;

        private JSONStorableBool isFocusWhenOnLock;
        private JSONStorableBool isOnlyFavorite;
        private JSONStorableBool isOpenPanelOnStart;
        private JSONStorableString saveStr;
        private JSONStorableFloat panelSize;

        private List<DAZBone> bonesList;
        private List<DAZMorph> dazMorphys;
        private List<DAZMorph> favMorphys;
        private DAZBones dazBones;
        private Canvas popupCanvas;
        private Camera targetCamera;
        private GameObject currentFloatPanel;
        private GameObject paneTemplate;
        private GameObject detailImg;
        private GameObject itemTemplateObj;
        private bool isSearchPose;

        private Text statusText;
        private Text linkInfoText;
        private Transform addBtnTrans;

        private string keywordFilterStr;
        private string regionFilterStr;
        private List<DAZMorph> morphyListToSelect;
        private Dictionary<string, GameObject> slidersUid2Obj = new Dictionary<string, GameObject>();
        private List<OptionData> optionDatas;

        private Dictionary<string, BoneMophyConfig> boneName2MorphyConfig = new Dictionary<string, BoneMophyConfig>();
        private List<DAZBone> morphyActiveBoneList = new List<DAZBone>();

        private AdsorbModule adsorbModule;
        private CameraHook cameraHook;
        private GLDraw glDraw;
        private HudHander hudHander;
        private MonitorCamModule monitorCamModule;
        private ScreenGrabModule screenGrabModule;
        private ClothModule clothModule;
        private MorphyCombineEditModule combineEditModule;
        private TimerManager timerManager;

        private string linkingLastUID;
        private string lastInfoOpenUID;
        private PDropDown dropdownMorphyList;

        // tip color
        private Color normalColor = new Color(50f / 255, 50f / 255, 50f / 255);
        private Color tiplColor = new Color(1, 0, 0);

        // main bones
        private string[] mainbones = new string[] {
            "hip","pelvis","chest","head","neck","rEye","lEye","rNipple","lNipple","rHand","lHand","lFoot","rFoot"
        };
        private List<DAZBone> mainBonesList = new List<DAZBone>();

        public GenerateDAZMorphsControlUI MorphControllerUI
        {
            get
            {
                JSONStorable storableByID = this.containingAtom.GetStorableByID("geometry");
                if (storableByID != null)
                {
                    DAZCharacterSelector dazcharacterSelector = storableByID as DAZCharacterSelector;
                    GenerateDAZMorphsControlUI morphsControlUI = dazcharacterSelector.morphsControlUI;
                    return morphsControlUI;
                }
                return null;
            }
        }

        public int CurrentOperateState
        {
            get { return currentOperateState; }
            set
            {
                currentOperateState = value;
                adsorbModule.IsAdsorbActive = value == OperateState.None ? false : true;
            }
        }

        public List<DAZBone> MorphyActiveBoneList
        {
            get
            {
                return morphyActiveBoneList;
            }
        }

        public List<DAZBone> BonesList
        {
            get
            {
                return this.bonesList;
            }
        }

        public List<DAZBone> MainBonesList
        {
            get {
                return this.mainBonesList;
            }
        }

        public GameObject CurrentFloatPanel
        {
            get { return currentFloatPanel; }
        }

        #region Load&&Save

        private void SaveAsPreset()
        {
            FileManagerSecure.CreateDirectory(presetSavePath);
            var fileBrowserUI = SuperController.singleton.fileBrowserUI;
            fileBrowserUI.SetTitle("Save EasyMorph Preset");
            fileBrowserUI.fileRemovePrefix = null;
            fileBrowserUI.hideExtension = false;
            fileBrowserUI.keepOpen = false;
            fileBrowserUI.fileFormat = presetEndfix;
            fileBrowserUI.defaultPath = presetSavePath;
            fileBrowserUI.showDirs = true;
            fileBrowserUI.shortCuts = null;
            fileBrowserUI.browseVarFilesAsDirectories = false;
            fileBrowserUI.SetTextEntry(true);
            fileBrowserUI.Show((path)=> {
                JSONClass jc = GetSaveJc();
                if (string.IsNullOrEmpty(path))
                    return;

                SaveJSON(jc, path);
                statusText.text = "Save Success!";
            });
            fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + presetEndfix;
            fileBrowserUI.ActivateFileNameField();
        }

        private void LoadAsPreset()
        {
            FileManagerSecure.CreateDirectory(presetSavePath);
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(presetSavePath);
            SuperController.singleton.GetMediaPathDialog((path)=> {
                if (string.IsNullOrEmpty(path))
                    return;

                var saveJsonNode = LoadJSON(path);
                LoadFromJc(saveJsonNode.ToString());
                statusText.text = "Load Success!";
            }, presetEndfix, presetSavePath, false, true, false, null, false, shortcuts);
        }
        
        private void OnAfterLoad()
        {
            string saveJson = saveStr.val;
            LoadFromJc(saveJson);
        }

        private void LoadFromJc(string saveJson)
        {
            JSONNode jc = JSONClass.Parse(saveJson);
            if (jc == null) return;

            boneName2MorphyConfig.Clear();
            combineEditModule.uid2LinkUids.Clear();
            morphyActiveBoneList.Clear();
            // restore custom config
            JSONArray bone2MorphConfig = jc["bone2MorphConfig"] as JSONArray;
            if (bone2MorphConfig.Count > 0)
            {
                foreach (var node in bone2MorphConfig.Childs)
                {
                    string key = node["boneName"].Value;
                    List<string> uids = node["morphs"].Childs.Select(x => x.Value).ToList();
                    if (!boneName2MorphyConfig.ContainsKey(key))
                    {
                        boneName2MorphyConfig.Add(key, new BoneMophyConfig() { uids = uids, boneName = key });
                    }

                    DAZBone bone = BonesList.FirstOrDefault(b => b.name == key);
                    morphyActiveBoneList.Add(bone);
                }
            }

            //[ {"morphID":[ "Breast Round" ]} ]
            JSONArray uid2linkConfig = jc["uid2linkConfig"] as JSONArray;
            if (uid2linkConfig.Count > 0)
            {
                foreach (var node in uid2linkConfig.Childs)
                {
                    string key = node["morphID"].Value;
                    if (!combineEditModule.uid2LinkUids.ContainsKey(key))
                    {
                        List<LinkedMorphData> linkedMorphDatas = new List<LinkedMorphData>();
                        node["uids"].Childs.ToList().ForEach((linkItem)=> {
                            LinkedMorphData dat = new LinkedMorphData() {
                                uid = linkItem["uid"].Value,
                                offset = float.Parse(linkItem["offset"].Value),
                                ralativeSpeed = float.Parse(linkItem["relativeSpeed"].Value)
                            };
                            linkedMorphDatas.Add(dat);
                        });
                        combineEditModule.uid2LinkUids.Add(key, linkedMorphDatas);
                    }
                }
            }
        }

        private void OnPreSave()
        {
            JSONClass jc = GetSaveJc();
            saveStr.val = jc.ToString();
        }

        private JSONClass GetSaveJc()
        {
            JSONClass jc = new JSONClass();

            // bone info
            //"bone2MorphConfig":[ {"boneName":"lNipple", "morphs":[ "Breast Height", "Breast Large" ]}
            if (boneName2MorphyConfig.Count > 0)
            {
                JSONArray bone2MorphConfig = new JSONArray();
                jc["bone2MorphConfig"] = bone2MorphConfig;

                boneName2MorphyConfig.Values.ToList().ForEach(config =>
                {
                    JSONClass temp = new JSONClass();
                    temp["boneName"] = config.boneName;

                    if (config.uids.Count > 0)
                    {
                        JSONArray tmpArray = new JSONArray();
                        temp["morphs"] = tmpArray;
                        config.uids.ForEach(uid =>
                        {
                            tmpArray.Add(uid);
                        });
                    }

                    bone2MorphConfig.Add(temp);
                });
            }

            // morph link info
            //{"boneName":"lEye", "morphs":[ "Eyes Closed" ]} ], "uid2linkConfig":[ {"morphID":"Eyes Closed", "uids":
            //[ {"uid":"Mouth Smile", "offset":"0", "relativeSpeed":"1.234121"}, {"uid":"Mouth Open", "offset":"0.5188598", "relativeSpeed":"0.1475532"} ]} ]
            if (combineEditModule.uid2LinkUids.Count > 0)
            {
                JSONArray uid2linkConfig = new JSONArray();
                jc["uid2linkConfig"] = uid2linkConfig;

                combineEditModule.uid2LinkUids.Keys.ToList().ForEach(key =>
                {
                    JSONClass temp = new JSONClass();
                    JSONArray tmpUids = new JSONArray();
                    temp["morphID"] = key;
                    temp["uids"] = tmpUids;

                    combineEditModule.uid2LinkUids[key].ForEach(linkItemData =>
                    {
                        JSONClass dat = new JSONClass();
                        dat["uid"] = linkItemData.uid;
                        dat["offset"] = linkItemData.offset.ToString();
                        dat["relativeSpeed"] = linkItemData.ralativeSpeed.ToString();
                        tmpUids.Add(dat);
                    });

                    uid2linkConfig.Add(temp);
                });
            }

            return jc;
        }

        #endregion

        public override void Init()
        {
            base.Init();

            var gfmaleRoot = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female");
            if (gfmaleRoot == null)
            {
                gfmaleRoot = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Male");
            }
            dazBones = gfmaleRoot.GetComponent<DAZBones>();
            bonesList = dazBones.dazBones.ToList();
            bonesList.ForEach(bone => {
                if (mainbones.Contains(bone.id))
                {
                    mainBonesList.Add(bone);
                }
            });
            
            targetCamera = SuperController.singleton.MonitorCenterCamera;
            currentGameMode = Utils.IsInVR() ? GameMode.VR : GameMode.Destop;

            MorphysPrepare();
            ModulesInit();
            UIInit();
            EventInit();

            // set oprate state
            CurrentOperateState = OperateState.None;
            adsorbModule.currentAdsorbMode = AdSorbMode.LinkMouse;
        }

        private void UIInit()
        {
            var assetsBundlePath = Utils.GetPluginPath(this) + "/resources/contex_panel_easymorphy.assetbundle";
            AssetLoader.AssetBundleFromFileRequest request = new AssetLoader.AssetBundleFromFileRequest { path = assetsBundlePath, callback = OnAssetBundleLoaded };
            AssetLoader.QueueLoadAssetBundleFromFile(request);

            popupCanvas = Utils.AppendCanvasObjectUnderHeadCenter("ContexViewSpaceCanvas", true);
            var canvasObj = popupCanvas.gameObject;

            isFocusWhenOnLock = Utils.SetupToggle(this, "IsFocusWhenOnLock", false, false);
            isOnlyFavorite = Utils.SetupToggle(this, "isOnlyFavorite", true, false);
            saveStr = Utils.SetupStorableString(this, "saveStr", "");
            isOpenPanelOnStart = Utils.SetupToggle(this, "isOpenUIOnStart", true, false);
            panelSize = Utils.SetupSliderFloat(this, "PanelSize", 1f, 0.3f, 1.5f, false);

            Utils.SetupButton(this, "ShowHide UI", () => {
                ShowHideEasyMorphPanel();
            }, false);

            Utils.SetupInfoText(this, "<size=40><b>EasyMorph 1.0</b></size>\n\n" +
                "<b>IsFocusWhenOnLock:</b> Whether to focus automatically or not.\n\n" +
                "<b>default shortCut:</b> Ctrl + H : Show/Hide main panel, Space : Change Operate Mode. you also can set custom shortCut by Keybinding Plugin.\n\n" +
                "<b>IsPriorityPrimaryNode : </b> cursor prioritize main joints like the head chest hips ankles and wrists for when assigning area morphs for posing\n\n" +
                "<b>Demo Video Link : </b> HangOut Range\n\n", 1000, true);

            // open panel on start
            timerManager.AddTimer(1, () => {
                if (isOpenPanelOnStart.val)
                {
                    CreateMorphPanel();
                }
            });
        }

        private void ModulesInit()
        {
            cameraHook = Utils.GetSafeComponent<CameraHook>(targetCamera.gameObject);
            cameraHook.onPostRenderCallback += OnPostRender;

            glDraw = new GLDraw();
            adsorbModule = new AdsorbModule(this);
            adsorbModule.Init();
            monitorCamModule = new MonitorCamModule(this);
            monitorCamModule.Init();
            screenGrabModule = new ScreenGrabModule(this);
            screenGrabModule.Init();
            clothModule = new ClothModule(this);
            clothModule.Init();
            timerManager = new TimerManager();
            combineEditModule = new MorphyCombineEditModule(this);
            combineEditModule.Init();
        }

        private void EventInit()
        {
            screenGrabModule.onCaptureATexture += OnCameraTextureGrabbed;
            SuperController.singleton.onBeforeSceneSaveHandlers += OnPreSave;
            SuperController.singleton.onSceneLoadedHandlers += OnAfterLoad;
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

            panelSize.setCallbackFunction = (size) => {
                currentFloatPanel.transform.localScale = Vector3.one * size;
            };
        }

        // texture grabbed
        private void OnCameraTextureGrabbed(Texture2D t2d)
        {
            SuperController.LogMessage(t2d.width.ToString());
        }

        private void Update()
        {
            adsorbModule.OnUpdate();
            screenGrabModule.OnUpdate();
            timerManager.Update();

            if (Input.GetKeyDown(KeyCode.Space))
            {
                StateSwichOperate();
            }

            if (Input.GetKey(KeyCode.LeftControl) && Input.GetKeyDown(KeyCode.H))
            {
                ShowHideEasyMorphPanel();
            }
        }

        private void StateSwichOperate()
        {
            // lock 
            adsorbModule.IsAdsorbLocked = !adsorbModule.IsAdsorbLocked;
            if (adsorbModule.IsAdsorbLocked)
            {
                // link line and focus
                adsorbModule.currentAdsorbMode = AdSorbMode.LinkUI;
                if (isFocusWhenOnLock.val)
                {
                    monitorCamModule.FocusPoint(adsorbModule.AdSorbData.point);
                }

                RefreshSlidersView();
            }
            // free
            else
            {
                adsorbModule.currentAdsorbMode = AdSorbMode.LinkMouse;
                ClearSliderObjects();
            }
        }

        private void RefreshSlidersView()
        {
            ClearSliderObjects();

            // refresh morphys map
            string mkey = adsorbModule.AdSorbData.dazBome.name;
            // create morphy item of the selected bone
            if (boneName2MorphyConfig.ContainsKey(mkey))
            {
                List<string> uids = boneName2MorphyConfig[mkey].uids;
                CreateSliderItemByUids(uids);
            }
        }

        private void ClearSliderObjects()
        {
            // clear last spawn objects
            List<GameObject> objs = slidersUid2Obj.Values.ToList();
            objs.ForEach(obj =>
            {
                Destroy(obj);
            });
            slidersUid2Obj.Clear();
        }

        private void OnAssetBundleLoaded(AssetLoader.AssetBundleFromFileRequest abffr)
        {
            try
            {
                var panelPrefab = abffr.assetBundle.LoadAsset<GameObject>("ContexPanel");
                paneTemplate = GameObject.Instantiate<GameObject>(panelPrefab, popupCanvas.transform);
                paneTemplate.transform.localRotation = Quaternion.identity;
                paneTemplate.transform.localScale = Vector3.one;
                paneTemplate.transform.localPosition = Vector3.zero;
                paneTemplate.SetActive(false);
            }
            catch (Exception e)
            {
                SuperController.LogError(e.Message);
                throw;
            }
        }

        private void ShowHideEasyMorphPanel()
        {
            if (currentFloatPanel == null)
            {
                CreateMorphPanel();
            }
            else
            {
                MorphPanleActive(!currentFloatPanel.activeInHierarchy);
            }
        }

        private void MorphPanleActive(bool isActive)
        {
            currentFloatPanel.SetActive(isActive);
            CurrentOperateState = isActive ? OperateState.AligningBoneNode : OperateState.None;
        }

        // create the actuall UI panel
        private void CreateMorphPanel()
        {
            if (currentFloatPanel != null) return;
            currentFloatPanel = GameObject.Instantiate<GameObject>(paneTemplate, popupCanvas.transform);

            RectTransform panelRectTrans = currentFloatPanel.GetComponent<RectTransform>();
            panelRectTrans.anchoredPosition = new Vector2(200, 200);
            currentFloatPanel.SetActive(true);

            // draggable
            currentFloatPanel.AddComponent<Draggable>();

            // ui init
            itemTemplateObj = currentFloatPanel.transform.Find("Scroll View/Viewport/Content/ItemTemplate").gameObject;
            addBtnTrans = currentFloatPanel.transform.Find("UpAnchor/BtnAdd");
            GameObject helpPanel = currentFloatPanel.transform.Find("PanelHelp").gameObject;
            Button addBtn = addBtnTrans.GetComponent<Button>();
            var keyworldinputFilter = currentFloatPanel.transform.Find("UpAnchor/keywordInputFilter").GetComponent<InputField>();
            var regionInputFilter = currentFloatPanel.transform.Find("UpAnchor/regionInputFilter").GetComponent<InputField>();
            dropdownMorphyList = currentFloatPanel.transform.Find("UpAnchor/DropdownMorphySelect").gameObject.AddComponent<PDropDown>();
            Toggle toggleFavorite = currentFloatPanel.transform.Find("UpAnchor/ToggleFavorite").GetComponent<Toggle>();
            detailImg = currentFloatPanel.transform.Find("DetailInfo").gameObject;
            Button btnSavePreset = currentFloatPanel.transform.Find("BtnSavePreset").GetComponent<Button>();
            Button btnLoadPreset = currentFloatPanel.transform.Find("BtnLoadPreset").GetComponent<Button>();
            Button btnHelp = currentFloatPanel.transform.Find("BtnHelp").GetComponent<Button>();
            Button btnClose = currentFloatPanel.transform.Find("BtnClose").GetComponent<Button>();
            Transform btnSearchTypeTrans = currentFloatPanel.transform.Find("UpAnchor/BtnSearchingType");
            Button btnSearchType = btnSearchTypeTrans.GetComponent<Button>();
            Text textSearchType = btnSearchTypeTrans.GetComponentInChildren<Text>();

            Toggle toggleShowHideCloth = currentFloatPanel.transform.Find("UpAnchor/ToggleShowHideCloth").GetComponent<Toggle>();

            statusText = currentFloatPanel.transform.Find("TextStatus").GetComponent<Text>();
            linkInfoText = currentFloatPanel.transform.Find("TextLinkInfo").GetComponent<Text>();

            // int ui state
            toggleFavorite.isOn = isOnlyFavorite.val;

            // event bind
            SyncDropdownList();

            btnClose.onClick.AddListener(() => {
                MorphPanleActive(false);
            });

            btnSearchType.onClick.AddListener(() => {
                isSearchPose = !isSearchPose;
                textSearchType.text = isSearchPose ? "Pose" : "Morph";
            });

            addBtn.onClick.AddListener(() =>
            {
                if (!adsorbModule.IsAdsorbLocked)
                {
                    SuperController.LogError("please select a bone anchor at first!");
                    return;
                }

                if (morphyListToSelect.Count == 0) return;

                var bone = adsorbModule.AdSorbData.dazBome;
                string uid = morphyListToSelect[dropdownMorphyList.Value].uid;
                if (bone != null)
                {
                    // normal mode
                    if (currentOperateState == OperateState.AligningBoneNode)
                    {
                        if (boneName2MorphyConfig.ContainsKey(bone.name))
                        {
                            List<string> uids = boneName2MorphyConfig[bone.name].uids;
                            if (!uids.Contains(uid))
                            {
                                uids.Add(uid);
                                CreateASliderObjectByUid(uid);
                            }
                            else
                            {
                                SuperController.LogError("aleady added a same morphy!");
                            }
                        }
                        else
                        {
                            boneName2MorphyConfig.Add(bone.name, new BoneMophyConfig()
                            {
                                boneName = bone.name,
                                uids = new List<string>() { uid }
                            });
                            morphyActiveBoneList.Add(bone);
                            CreateASliderObjectByUid(uid);
                        }
                    }
                    // linking
                    else if (currentOperateState == OperateState.Linking)
                    {
                        if (combineEditModule.uid2LinkUids.ContainsKey(linkingLastUID))
                        {
                            combineEditModule.uid2LinkUids[linkingLastUID].Add(new LinkedMorphData() {
                                 uid = uid
                            });
                        }
                        else
                        {
                            combineEditModule.uid2LinkUids.Add(linkingLastUID, new List<LinkedMorphData>() {
                                 new LinkedMorphData(){
                                      uid = uid
                                 }
                            });
                        }
                        linkInfoText.text = string.Format("Success Link: {0}", Utils.GetSimplestMorphName(uid));
                    }
                }

            });

            btnLoadPreset.onClick.AddListener(() => {
                LoadAsPreset();
            });

            btnSavePreset.onClick.AddListener(() => {
                SaveAsPreset();
            });

            btnHelp.onClick.AddListener(() => {
                helpPanel.gameObject.SetActive(!helpPanel.activeInHierarchy);
            });

            keyworldinputFilter.onEndEdit.AddListener((str) =>
            {
                keywordFilterStr = str;
                SyncDropdownList();
                dropdownMorphyList.DropListActive = false;
            });

            regionInputFilter.onEndEdit.AddListener((str) =>
            {
                regionFilterStr = str;
                SyncDropdownList();
                dropdownMorphyList.DropListActive = false;
            });

            toggleFavorite.onValueChanged.AddListener((val) => {
                isOnlyFavorite.val = val;
            });

            toggleShowHideCloth.onValueChanged.AddListener((val) => {
                clothModule.ShowHideAllCloth(val);
            });

            dropdownMorphyList.onSelectValueChanged = (index, go) => {

            };

            // state set
            adsorbModule.currentAdsorbMode = AdSorbMode.LinkMouse;
            currentOperateState = OperateState.AligningBoneNode;
            adsorbModule.IsAdsorbActive = true;
        }

        private void SyncDropdownList()
        {
            morphyListToSelect = new List<DAZMorph>();
            List<DAZMorph> sourceMorphys = isOnlyFavorite.val ? this.favMorphys : dazMorphys;
            optionDatas = new List<OptionData>();
            // filter out the select items
            if (sourceMorphys.Count > 0)
            {
                sourceMorphys.ForEach(mor =>
                {
                    if ((isSearchPose && mor.isPoseControl) || (isSearchPose == false && mor.isPoseControl == false))
                    {
                        if (string.IsNullOrEmpty(regionFilterStr))
                        {
                            FilterMorphyWithKeyword(mor);
                        }
                        // region only need to contain
                        else if (mor.region.ToLower().Contains(regionFilterStr.ToLower()))
                        {
                            FilterMorphyWithKeyword(mor);
                        }
                    }
                });

                // update dropdown info
                Text firstText = dropdownMorphyList.transform.Find("Label").GetComponent<Text>();
                firstText.text = "";
                if (optionDatas.Count > 0)
                {
                    
                    firstText.text = optionDatas[0].morphName;
                }
            }
            dropdownMorphyList.Options = optionDatas;
        }

        private void FilterMorphyWithKeyword(DAZMorph mor)
        {
            if (string.IsNullOrEmpty(keywordFilterStr) || mor.resolvedDisplayName.ToLower().Contains(keywordFilterStr.ToLower()))
            {
                morphyListToSelect.Add(mor);
                optionDatas.Add(new OptionData() { morphName = mor.resolvedDisplayName, region = mor.region });
            }
        }

        private void CreateSliderItemByUids(List<string> uids)
        {
            if (uids.Count == 0) return;

            uids.ForEach(uid => {
                CreateASliderObjectByUid(uid);
            });
        }

        private void CreateASliderObjectByUid(string uid)
        {
            Transform itemParent = itemTemplateObj.transform.parent;
            var morphy = MorphControllerUI.GetMorphByUid(uid);
            var test = Instantiate<GameObject>(itemTemplateObj, itemParent).transform;
            slidersUid2Obj.Add(uid, test.gameObject);
            test.gameObject.name = uid;
            test.gameObject.SetActive(true);

            SliderForm(morphy, test);
        }

        private void SliderForm(DAZMorph morphy, Transform test)
        {
            if (test != null)
            {
                // 设置UI
                Slider slider = test.Find("SliderValue").GetComponent<Slider>();
                Text textName = test.Find("TextParaName").GetComponent<Text>();
                Text textMorphType = test.Find("TextMorphType").GetComponent<Text>();
                Button btnDefaut = test.Find("BtnDefault").GetComponent<Button>();
                Button btnMin = test.Find("BtnMin").GetComponent<Button>();
                Button btnMax = test.Find("BtnMax").GetComponent<Button>();
                InputField valueInput = test.Find("ValueInputField").GetComponent<InputField>();
                Transform btnlinkTrans = test.Find("BtnLink");
                Button btnLink = btnlinkTrans.GetComponent<Button>();
                Button btnRemove = test.Find("BtnRemove").GetComponent<Button>();
                Transform btnInfoTrans = test.Find("BtnInfo");
                Button btnInfo = btnInfoTrans.GetComponent<Button>();
                GameObject btnInfoReddot = btnInfoTrans.Find("RedDot").gameObject;

                Text lb = btnlinkTrans.GetComponentInChildren<Text>();
                Text addlb = addBtnTrans.GetComponentInChildren<Text>();

                // 同步morphy数据
                textName.text = morphy.resolvedDisplayName;
                slider.minValue = morphy.min;
                slider.maxValue = morphy.max;
                slider.value = morphy.morphValue;
                valueInput.text = slider.value.ToString();

                textMorphType.text = morphy.isPoseControl ? "POSE" : "MORPH";

                btnMin.onClick.AddListener(() =>
                {
                    slider.value = morphy.min;
                });

                btnMax.onClick.AddListener(() =>
                {
                    slider.value = morphy.max;
                });

                btnRemove.onClick.AddListener(() =>
                {
                    var bone = adsorbModule.AdSorbData.dazBome;
                    string uid = test.name;

                    List<string> sliderUids = boneName2MorphyConfig[bone.name].uids;
                    sliderUids.Remove(uid);

                    // clear bone square mark
                    if (sliderUids.Count == 0)
                    {
                        morphyActiveBoneList.Remove(bone);
                        boneName2MorphyConfig.Remove(bone.name);
                    }

                    slidersUid2Obj.Remove(uid);
                    Destroy(test.gameObject);

                    // if linking, reset operate state
                    if (currentOperateState == OperateState.Linking && linkingLastUID == test.name)
                    {
                        SwitchOperateState(lb, addlb);
                    }
                });

                // link other morph
                btnLink.onClick.AddListener(() =>
                {;

                    if (currentOperateState == OperateState.Linking)
                    {
                        if (linkingLastUID == test.name)
                        {
                            SwitchOperateState(lb, addlb);
                        }
                    }
                    else
                    {
                        SwitchLinkingState(test, lb, addlb);
                    }
                });

                valueInput.onEndEdit.AddListener((str) =>
                {
                    slider.value = float.Parse(str);
                });

                btnInfo.onClick.AddListener(() => {
                    combineEditModule.OpenMorphBindPanel(detailImg.transform, test.name);
                    lastInfoOpenUID = test.name;
                });

                // default btn
                btnDefaut.onClick.AddListener(() =>
                {
                    slider.value = morphy.jsonFloat.defaultVal;
                });

                // set value
                slider.onValueChanged.AddListener((v) =>
                {
                    valueInput.text = v.ToString();
                    morphy.morphValue = v;

                    // sync morphlink
                    if (combineEditModule.uid2LinkUids.ContainsKey(test.name))
                    {
                        List<LinkedMorphData> linkUids = combineEditModule.uid2LinkUids[test.name];
                        linkUids.ForEach(x =>
                        {
                            var tmpMorph = MorphControllerUI.GetMorphByUid(x.uid);
                            float targetValue = x.offset + morphy.morphValue * x.ralativeSpeed;
                            targetValue = Mathf.Clamp(targetValue, tmpMorph.min, tmpMorph.max);
                            tmpMorph.morphValue = targetValue;
                        });
                    }
                });

                // reddot
                btnInfoReddot.SetActive(combineEditModule.uid2LinkUids.ContainsKey(test.name) 
                    && combineEditModule.uid2LinkUids[test.name].Count > 0);
            }
        }

        private void SwitchLinkingState(Transform test, Text lb, Text addlb)
        {
            lb.text = "Linking...";
            lb.color = tiplColor;
            statusText.text = "On Linking Mode...";
            linkInfoText.text = "";
            statusText.color = tiplColor;
            linkingLastUID = test.name;
            currentOperateState = OperateState.Linking;
            addlb.color = tiplColor;
        }

        private void SwitchOperateState(Text lb, Text addlb)
        {
            currentOperateState = OperateState.AligningBoneNode;
            lb.text = "Link";
            lb.color = normalColor;
            statusText.text = "On Normal Mode...";
            linkInfoText.text = "";
            statusText.color = normalColor;
            addlb.color = normalColor;
        }

        private void MorphysPrepare()
        {
            JSONStorable storableByID = this.containingAtom.GetStorableByID("geometry");
            if (storableByID != null)
            {
                DAZCharacterSelector dazcharacterSelector = storableByID as DAZCharacterSelector;
                GenerateDAZMorphsControlUI morphsControlUI = dazcharacterSelector.morphsControlUI;
                if (morphsControlUI != null)
                {
                    this.dazMorphys = morphsControlUI.GetMorphs();
                    this.favMorphys = dazMorphys.Where(x => x.favorite == true).ToList();
                }
            }
        }

        private void OnPostRender()
        {
            if (currentGameMode == GameMode.Destop)
            {
                adsorbModule.OnPostRender();
                screenGrabModule.OnPostRender();
            }
        }

        private void OnApplicationFocus(bool focus)
        {
            screenGrabModule.OnAppFocus(focus);
        }

        // keybinding
        public void OnBindingsListRequested(List<object> bindings)
        {
            bindings.Add(new JSONStorableAction("ShowHidePanel", ShowHideEasyMorphPanel));
            bindings.Add(new JSONStorableAction("ApplySet", StateSwichOperate));
            bindings.Add(new JSONStorableAction("ShowHideMarkPoint", ShowHideMarkPoint));
        }

        private void ShowHideMarkPoint()
        {
            adsorbModule.isDrawMarkPoint = !adsorbModule.isDrawMarkPoint;
        }

        private void OnDestroy()
        {
            cameraHook.onPostRenderCallback -= OnPostRender;
            adsorbModule.UnInit();
            monitorCamModule.UnInit();
            screenGrabModule.UnInit();
            clothModule.UnInit();
            combineEditModule.UnInit();
            screenGrabModule.onCaptureATexture -= OnCameraTextureGrabbed;
            SuperController.singleton.onBeforeSceneSaveHandlers -= OnPreSave;
            SuperController.singleton.onSceneLoadedHandlers -= OnAfterLoad;
            SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            Destroy(popupCanvas.gameObject);
        }

    }
}
