﻿using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using UnityEngine.UI;
using System;

namespace PluginIdea
{
    public class LinkedMorphData
    {
        public string uid;
        public float offset = 0;
        public float ralativeSpeed = 1;
    }

    public class MorphyCombineEditModule : EasyMorphyModuleBase
    {

        struct RecordStep
        {
            public const int None = 0;
            public const int ReadyTo = 1;
            public const int Step1 = 2;
            public const int Step2 = 3;
            public const int AfterFinished = 4;
        }

        class RecordData
        {
            public float backupOffset;
            public float start;
            public Slider slider;
        }

        public Dictionary<string, List<LinkedMorphData>> uid2LinkUids = new Dictionary<string, List<LinkedMorphData>>();
        private List<GameObject> linkInfoObjects = new List<GameObject>();
        private Dictionary<string, RecordData> uid2RmpRecord = new Dictionary<string, RecordData>();

        private Transform rootTrans;
        private string mainMorphUid;
        private GameObject linkInfoItemTemplate;
        private GenerateDAZMorphsControlUI morphControlUI;
        private Slider btnSliderMainMorph;

        private int currentRecordStep = RecordStep.None;

        private float mainMorphMapStartValue, mainMorphMapEndValue;

        private Color normalColor = new Color(50f / 255, 50f / 255, 50f / 255);
        private Color tiplColor = new Color(1, 0, 0);

        public MorphyCombineEditModule(EasyMorphy ower)
        {
            this.ower = ower;
        }

        public override void Init()
        {
            this.morphControlUI = ower.MorphControllerUI;
        }

        private void InitUI()
        {
            linkInfoItemTemplate = rootTrans.Find("Template/Viewport/Content/Item").gameObject;

            Transform btnRecordTrans = rootTrans.Find("BtnRecordData");
            Button btnRecord = btnRecordTrans.GetComponentInChildren<Button>();
            Text text = btnRecordTrans.GetComponentInChildren<Text>();
            Button btnClose = rootTrans.Find("BtnDetailClose").GetComponentInChildren<Button>();

            // init
            currentRecordStep = RecordStep.None;
            text.color = normalColor;
            text.text = "Record";

            // base info
            SetMainMorphSlider();

            // uievent
            btnClose.onClick.RemoveAllListeners();
            btnClose.onClick.AddListener(() => {
                rootTrans.gameObject.SetActive(false);

                // cancel the changes when record operate not finished
                if (currentRecordStep != RecordStep.None)
                {
                    if (uid2LinkUids.ContainsKey(mainMorphUid))
                    {
                        List<LinkedMorphData> linkUids = uid2LinkUids[mainMorphUid];
                        linkUids.ForEach(x =>
                        {
                            if (uid2RmpRecord.ContainsKey(x.uid))
                            {
                                x.offset = uid2RmpRecord[x.uid].backupOffset;
                            }
                        });
                    }
                }
            });

            btnRecord.onClick.RemoveAllListeners();
            btnRecord.onClick.AddListener(() => {

                currentRecordStep++;
                if (currentRecordStep == RecordStep.ReadyTo)
                {
                    text.color = tiplColor;
                    text.text = "mark Min";

                    if (uid2LinkUids.ContainsKey(mainMorphUid))
                    {
                        List<LinkedMorphData> linkUids = uid2LinkUids[mainMorphUid];
                        linkUids.ForEach(x =>
                        {
                            if (uid2RmpRecord.ContainsKey(x.uid))
                            {
                                uid2RmpRecord[x.uid].backupOffset = x.offset;
                            }
                        });
                    }
                }
                else if (currentRecordStep == RecordStep.Step1)
                {
                    mainMorphMapStartValue = btnSliderMainMorph.value;
                    text.text = "mark Max";

                    if (uid2LinkUids.ContainsKey(mainMorphUid))
                    {
                        List<LinkedMorphData> linkUids = uid2LinkUids[mainMorphUid];
                        linkUids.ForEach(x =>
                        {
                            if (uid2RmpRecord.ContainsKey(x.uid))
                            {
                                Slider tmpSlider = uid2RmpRecord[x.uid].slider;
                                x.offset = tmpSlider.value - btnSliderMainMorph.value;
                                x.offset = Mathf.Clamp(x.offset, -2, 2);

                                uid2RmpRecord[x.uid].start = tmpSlider.value;
                            }
                        });
                    }
                }
                else if (currentRecordStep == RecordStep.Step2)
                {
                    mainMorphMapEndValue = btnSliderMainMorph.value;
                    text.text = "Apply?";

                    if (uid2LinkUids.ContainsKey(mainMorphUid))
                    {
                        List<LinkedMorphData> linkUids = uid2LinkUids[mainMorphUid];
                        linkUids.ForEach(x =>
                        {
                            if (uid2RmpRecord.ContainsKey(x.uid))
                            {
                                Slider tmpSlider = uid2RmpRecord[x.uid].slider;
                                float mainLeng = mainMorphMapEndValue - mainMorphMapStartValue;
                                float slLen = tmpSlider.value - uid2RmpRecord[x.uid].start;
                                x.ralativeSpeed = slLen / mainLeng;
                                x.ralativeSpeed = Mathf.Clamp(x.ralativeSpeed, 0, 100);
                            }
                        });
                    }
                }
                else if (currentRecordStep == RecordStep.AfterFinished)
                {
                    currentRecordStep = RecordStep.None;
                    text.text = "Record";
                    text.color = normalColor;
                }
            });
        }

        private void SetMainMorphSlider()
        {
            btnSliderMainMorph = rootTrans.Find("SliderRootMorph").GetComponentInChildren<Slider>();
            InputField inputMainMorphValue = rootTrans.Find("InputRootMorphValue").GetComponentInChildren<InputField>();
            Text RootMorphName = rootTrans.Find("RootMorphName").GetComponentInChildren<Text>();

            // set value
            var mainMorph = morphControlUI.GetMorphByUid(mainMorphUid);
            btnSliderMainMorph.minValue = mainMorph.min;
            btnSliderMainMorph.maxValue = mainMorph.max;
            btnSliderMainMorph.value = mainMorph.morphValue;
            RootMorphName.text = mainMorphUid;
            inputMainMorphValue.text = btnSliderMainMorph.value.ToString();

            // event
            inputMainMorphValue.onEndEdit.RemoveAllListeners();
            inputMainMorphValue.onEndEdit.AddListener((str) => {
                btnSliderMainMorph.value = float.Parse(str);
                mainMorph.SetValue(btnSliderMainMorph.value);
            });

            btnSliderMainMorph.onValueChanged.RemoveAllListeners();
            btnSliderMainMorph.onValueChanged.AddListener((v) => {
                inputMainMorphValue.text = v.ToString();
                mainMorph.SetValue(v);

                if (currentRecordStep != RecordStep.None) return;

                // sync linkedUids
                if (uid2LinkUids.ContainsKey(mainMorphUid))
                {
                    List<LinkedMorphData> linkUids = uid2LinkUids[mainMorphUid];
                    linkUids.ForEach(x =>
                    {
                        if (uid2RmpRecord.ContainsKey(x.uid))
                        {
                            Slider tmpSlider = uid2RmpRecord[x.uid].slider;
                            float targetValue = x.offset + btnSliderMainMorph.value * x.ralativeSpeed;
                            targetValue = Mathf.Clamp(targetValue, tmpSlider.minValue, tmpSlider.maxValue);
                            tmpSlider.value = targetValue;
                            var tmpMorph = morphControlUI.GetMorphByUid(x.uid);
                            tmpMorph.SetValue(targetValue);
                        }
                    });
                }
            });
        }

        public void OpenMorphBindPanel(Transform panelTrans, string mainMorphUid)
        {
            this.rootTrans = panelTrans;
            this.mainMorphUid = mainMorphUid;

            InitUI();

            // clear
            uid2RmpRecord.Clear();
            if (linkInfoObjects.Count > 0)
            {
                linkInfoObjects.ForEach(obj => {
                    GameObject.Destroy(obj);
                });
                linkInfoObjects.Clear();
            }

            // create items
            if (uid2LinkUids.ContainsKey(mainMorphUid))
            {
                uid2LinkUids[mainMorphUid].ForEach(linkdata => {
                    GameObject tmp = GameObject.Instantiate<GameObject>(linkInfoItemTemplate, linkInfoItemTemplate.transform.parent);

                    InputField inputValue = tmp.transform.Find("InputValue").GetComponent<InputField>();
                    Text lable = tmp.transform.Find("Item Label").GetComponent<Text>();
                    Button btnRemove = tmp.transform.Find("BtnRemove").GetComponent<Button>();
                    Slider sliderValue = tmp.transform.Find("SliderValue").GetComponent<Slider>();

                    if (!uid2RmpRecord.ContainsKey(linkdata.uid)) uid2RmpRecord.Add(linkdata.uid, new RecordData() { slider = sliderValue });

                    // data
                    var tmpMorph = morphControlUI.GetMorphByUid(linkdata.uid);
                    lable.text = linkdata.uid;
                    inputValue.text = tmpMorph.morphValue.ToString() ;
                    sliderValue.value = tmpMorph.morphValue;

                    // set
                    linkInfoObjects.Add(tmp);

                    // event
                    btnRemove.onClick.AddListener(() => {
                        uid2LinkUids[mainMorphUid].Remove(linkdata);
                        if (uid2RmpRecord.ContainsKey(linkdata.uid)) uid2RmpRecord.Remove(linkdata.uid);
                        GameObject.Destroy(tmp);
                    });

                    sliderValue.onValueChanged.AddListener((val) => {
                        inputValue.text = val.ToString();
                        tmpMorph.SetValue(val);
                    });

                    tmp.name = linkdata.uid;
                    tmp.SetActive(true);
                });
            }

            rootTrans.gameObject.SetActive(true);
        }

        public override void UnInit()
        {
            
        }
    }
}

