﻿using UnityEngine;
using System.Collections;
using System;

namespace PluginIdea
{
    public class ScreenGrabModule : EasyMorphyModuleBase
    {

        public Action<Texture2D> onCaptureATexture;

        struct SelectState
        {
            public const int NONE = 0;
            public const int BOXSELECT_READY = 1;
        }

        private GLDraw glDraw;
        private Color rectColor;
        private Vector2 start, end;
        private bool isOnDrag;
        private int selectState;

        public ScreenGrabModule(EasyMorphy ower)
        {
            this.ower = ower;
        }

        public override void Init()
        {
            this.glDraw = new GLDraw();
            this.rectColor = new Color(0, 1, 0, 0.05f);
            this.selectState = SelectState.NONE;
        }

        public void OnAppFocus(bool focus)
        {
            if (focus)
            {
                selectState = SelectState.NONE;
            }
        }

        public override void OnUpdate()
        {
            if (selectState == SelectState.NONE)
            {
                // box select
                bool lalt = Input.GetKey(KeyCode.LeftAlt);
                if (lalt)
                {
                    selectState = SelectState.BOXSELECT_READY;
                }
            }

            if (Input.GetKeyUp(KeyCode.LeftAlt))
            {
                selectState = SelectState.NONE;
            }
        }

        private Rect captureRect;
        public override void OnPostRender()
        {
            if (selectState == SelectState.BOXSELECT_READY)
            {
                // draw box
                if (Input.GetMouseButton(0))
                {
                    if (!isOnDrag)
                    {
                        isOnDrag = true;
                        start = new Vector2(Input.mousePosition.x, Input.mousePosition.y);
                    }
                    end = new Vector2(Input.mousePosition.x, Input.mousePosition.y);

                    captureRect = new Rect(start, new Vector2(end.x - start.x, end.y - start.y));
                    glDraw.DrawQuad2D(captureRect, rectColor);
                }

                if (Input.GetMouseButtonUp(0))
                {
                    isOnDrag = false;
                    //Texture2D screenShot = CaptureCamera(SuperController.singleton.MonitorCenterCamera, captureRect);
                    //if (onCaptureATexture != null) onCaptureATexture(screenShot);
                }
            }
        }

        // capture texture
        private Texture2D CaptureCamera(Camera camera, Rect rect)
        {
            RenderTexture rt = new RenderTexture((int)rect.width, (int)rect.height, 0);
            camera.targetTexture = rt;
            camera.Render();
            // 激活这个rt, 并从中中读取像素。
            RenderTexture.active = rt;
            Texture2D screenShot = new Texture2D((int)rect.width, (int)rect.height, TextureFormat.RGB24, false);
            screenShot.ReadPixels(rect, 0, 0);// 注：这个时候，它是从RenderTexture.active中读取像素
            screenShot.Apply();

            // 重置相关参数，以使用camera继续在屏幕上显示
            camera.targetTexture = null;
            RenderTexture.active = null; // JC: added to avoid errors
            GameObject.Destroy(rt);
            return screenShot;
        }

        public override void UnInit()
        {
            
        }
    }
}


